// This demo application performs continuous background subtraction and broadcasts bounding boxes of foreground objects.

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>
#include "cameraboard.h"

// maximum number of blobs per frame
#define MAX_BLOBS		128

#define BASESTATION_ID 0

#define diff(x,y) ((x>y)?(x-y):(y-x))

/**
	mbbox contains the bounding boxes found in each frame.  the index I for each array is valid for I<count.
		count - number of bounding boxes in the frame
		xmin - minimum bound of the box along the x axis
		xmax - maximum bound of the box along the x axis
		ymin - minimum bound of the box along the y axis
		ymax - maximum bound of the box along the y axis
*/
typedef struct mbbox {
	int count;
	int xmin[MAX_BLOBS];
	int xmax[MAX_BLOBS];
	int ymin[MAX_BLOBS];
	int ymax[MAX_BLOBS];
} mbbox;

/**
	bgmodel contains the all data and parameters relevant to background subtraction.
		width, height - width and height of the image
		length - equivalent to width*height
		count - frame counter
		blend - factor used to determine the blending factor between fg and mean when updating the mean
		thresh - threshold value used to classify pixels as background or foreground
		sqsize - square size used in median filtering
		pixel_distance - image array that contains the absolute difference between mean and current frame
		mean - image array containing the running mean of the background
		fg - image array containing the calculated foreground for the current frame
		bg - image array containing the calculated background for the current frame
		fgmask - image array containing the foreground binary mask
		mbb - pointer to the mbbox cointaining bounding box information for the current frame
*/
typedef struct bgmodel {
	int width, height;
	int length;
	int count;
	int blend; // determines blending factor between mean and fg
	int thresh;
	int sqsize; // for smoothing fg mask
	unsigned char* pixel_distance;
	unsigned char* mean;
	unsigned char* fg;
	unsigned char* bg;
	unsigned char* fgmask;
	mbbox* mbb;
} bgmodel;

/**
	smooth_median performs median filtering on BGM->fgmask, using the parameters contained in BGM
		bgm - pointer to the bgmodel.  median filtering is destructively performed on the image array bgm->fgmask
		
*/
void smooth_median(bgmodel* bgm) {
	int w = bgm->width;
	int h = bgm->height;
	int i, j;
	int k, l;

	int sqoffset = bgm->sqsize/2; // sqsize/2 rounded down, ie. sqsize=5, sqoffset=2

	unsigned char* tempmask = (unsigned char*) malloc(h*w*sizeof(unsigned char));
	
	// set outside border of mask to 0
	for(i=0; i<sqoffset; i++) { // top edge
		for(j=0; j<w; j++) {
			tempmask[i*w+j] = 0;
		}
	}
	for(i=sqoffset; i<(h-sqoffset); i++) { // left edge
		for(j=0; j<sqoffset; j++) {
			tempmask[i*w+j] = 0;	
		}
	}
	for(i=sqoffset; i<(h-sqoffset); i++) { // right edge
		for(j=(w-sqoffset); j<w; j++) {
			tempmask[i*w+j] = 0;	
		}
	}
	for(i=(h-sqoffset); i<h; i++) { // bottom edge
		for(j=0; j<w; j++) {
			tempmask[i*w+j] = 0;
		}
	}

	for (i=sqoffset; i<(h-sqoffset); i++) {
		for (j=sqoffset; j<(w-sqoffset); j++) {
			int sum = 0;
			for (k=0; k<bgm->sqsize; k++) {
				for (l=0; l<bgm->sqsize; l++) {
					int ypos = i+k-sqoffset;
					int xpos = j+l-sqoffset;
					if ((xpos>=0) && (xpos<w) && (ypos>=0) && (ypos<h)) {
						sum += (bgm->fgmask[ypos*w+xpos] > 0);
					}
				}
			}

			if ((bgm->sqsize*bgm->sqsize-sum*2)<0) { // mask pixel
				tempmask[i*w+j] = 255;
			} else { // no mask pixel
				tempmask[i*w+j] = 0;
			}
		}
	}

	free(bgm->fgmask);
	bgm->fgmask = tempmask;
}

void get_mbb(bgmodel* bgm, int blob_size_thresh) {
	int i,j,k,l;
	int blobx, bloby;
	
	int visited[bgm->length];
	int blob[bgm->length];
	int blob_length = 0;
	
	int xmin, xmax, ymin, ymax;
	
	bgm->mbb->count = 0;
	
	// init visited
	for (i=0; i<bgm->length; i++) {
		visited[i] = 0;
	}
	
	for (i=0; i<bgm->length; i++) {
		if (!visited[i] && bgm->fgmask[i]) { // new blob detected
			// init blob vector
			blob[0] = i;
			blob_length = 1;
			visited[i] = 1;
			
			// check left
			if ((i%bgm->width) != 0) { // not leftmost column
				j = i-1;
				if (!visited[j] && bgm->fgmask[j]) { // grow blob
					blob[blob_length++] = j;
					visited[j] = 1;
				}
			}
			// check up
			if (i >= bgm->width) { // not top row pixel
				j = i-bgm->width;
				if (!visited[j] && bgm->fgmask[j]) { // grow blob
					blob[blob_length++] = j;
					visited[j] = 1;
				}
			}
			// check right
			if ((i%bgm->width) != (bgm->width-1)) { // not rightmost column
				j = i+1;
				if (!visited[j] && bgm->fgmask[j]) { // grow blob
					blob[blob_length++] = j;
					visited[j] = 1;
				}
			}
			// check down
			if (i < (bgm->length-bgm->width)) { // not rightmost column
				j = i+bgm->width;
				if (!visited[j] && bgm->fgmask[j]) { // grow blob
					blob[blob_length++] = j;
					visited[j] = 1;
				}
			}
			
			// begin growing blob
			j = 1;
			while (j < blob_length) {
				k = blob[j]; // next pixel to check for growth
				
				// check left
				if ((k%bgm->width) != 0) { // not leftmost column
					l = k-1;
					if (!visited[l] && bgm->fgmask[l]) { // grow blob
						blob[blob_length++] = l;
						visited[l] = 1;
					}
				}
				// check up
				if (k >= bgm->width) { // not top row pixel
					l = k-bgm->width;
					if (!visited[l] && bgm->fgmask[l]) { // grow blob
						blob[blob_length++] = l;
						visited[l] = 1;
					}
				}
				// check right
				if ((k%bgm->width) != (bgm->width-1)) { // not rightmost column
					l = k+1;
					if (!visited[l] && bgm->fgmask[l]) { // grow blob
						blob[blob_length++] = l;
						visited[l] = 1;
					}
				}
				// check down
				if (k < (bgm->length-bgm->width)) { // not rightmost column
					l = k+bgm->width;
					if (!visited[l] && bgm->fgmask[l]) { // grow blob
						blob[blob_length++] = l;
						visited[l] = 1;
					}
				}
				
				j++;
			}
			
			// add blob and get bounding box
			if (blob_length > blob_size_thresh) {
				// init bounding box
				xmin = bgm->width-1;
				xmax = 0;
				ymin = bgm->height-1;
				ymax = 0;
				
				// get bounding box
				for (j=0; j<blob_length; j++) {
					blobx = blob[j]%bgm->width;
					bloby = blob[j]/bgm->width;
					
					if (blobx < xmin) xmin = blobx;
					if (blobx > xmax) xmax = blobx;
					if (bloby < ymin) ymin = bloby;
					if (bloby > ymax) ymax = bloby;
				}
				
				// add bounding box
				bgm->mbb->xmin[bgm->mbb->count] = xmin;
				bgm->mbb->xmax[bgm->mbb->count] = xmax;
				bgm->mbb->ymin[bgm->mbb->count] = ymin;
				bgm->mbb->ymax[bgm->mbb->count] = ymax;
				bgm->mbb->count++;
				
				printf("blob size = %8d, x %5d %5d, y %5d %5d\n", blob_length, xmin, xmax, ymin, ymax);
			} else { // blob too small, delete from mask
				for (j=0; j<blob_length; j++) {
					bgm->fgmask[blob[j]] = 0;
				}
			}
		}
	}
}

/**
	init_bgmodel initializes a bgmodel struct including allocating memory space.  returns the newly allocated and initialized bgmodel
		img - image array of the current frame.  the initial mean background is set to this
		width - width of IMG
		height - height of IMG
		thresh - threshold value used for background subtraction
		sqsize - square size used for median filtering
		blend - blend value used to blend the current frame and the running mean
*/
bgmodel* init_bgmodel(unsigned char* img, int width, int height, int thresh, int sqsize, int blend) {
	int i;
	bgmodel* bgm = malloc(sizeof(bgmodel));

	// init len and count
	bgm->width = width;
	bgm->height = height;
	bgm->length = width*height;
	bgm->count = 1;

	bgm->thresh = thresh;
	bgm->sqsize = sqsize;
	bgm->blend = blend;

	// init pixel_distance
	bgm->pixel_distance = (unsigned char*) malloc(bgm->length*sizeof(unsigned char));
	
	// init mean
	bgm->mean = (unsigned char*) malloc(bgm->length*sizeof(unsigned char));
	memcpy(bgm->mean, img, bgm->length);

	// init fg and bg
	bgm->fg = (unsigned char*) malloc(bgm->length*sizeof(unsigned char));
	bgm->bg = (unsigned char*) malloc(bgm->length*sizeof(unsigned char));

	// init fgmask
	bgm->fgmask = (unsigned char*) malloc(bgm->length*sizeof(unsigned char));
	
	// init mbbox
	bgm->mbb = malloc(sizeof(mbbox));
	bgm->mbb->count = 0;

	return bgm;
}

/**
	update_bgmodel updates the bgmodel BGM with the current frame IMG.  after update_bgmodel has executed, BGM will contain the background, foreground, and bounding box information corresponding to the current frame IMG
		bgm - pointer to the bgmodel.  update_bgmodel destructively modifies bgm
		img - image array containing the current frame
*/
void update_bgmodel(bgmodel* bgm, unsigned char* img) {
	int i,j;
	int temp_mean;
	
	// calculate fg mask, update bg and bg mean
	for (i=0; i<bgm->length; i++) {
		// abs diff + thresh
		bgm->pixel_distance[i] = abs(bgm->mean[i] - img[i]);
		j = (bgm->pixel_distance[i] > bgm->thresh);
		if (j)
			bgm->fgmask[i] = 255;
		else
			bgm->fgmask[i] = 0;
		
		// update bg and bg mean
		if (j) { // pixel is fg
			bgm->bg[i] = bgm->mean[i];
		} else { // pixel is bg
			bgm->bg[i] = img[i];
		}
		// update bg mean
		temp_mean = (bgm->mean[i]*(bgm->blend-1) + img[i]) / bgm->blend;
		if ((img[i] > bgm->mean[i]) && (bgm->mean[i] == temp_mean))
			bgm->mean[i] = bgm->mean[i]+1;
		else
			bgm->mean[i] = temp_mean;
	}
	
	//writeFileBinary(bgm->fgmask, bgm->width, bgm->height, "ma", 0);
	
	// smooth fg mask
	smooth_median(bgm);
	
	// get bounding boxes, update fgmask
	get_mbb(bgm, (bgm->length/80));
	
	// update fg
	for (i=0; i<bgm->length; i++) {
		bgm->fg[i] = bgm->fgmask[i] * img[i];
	}

	// increment count
	bgm->count++;
}

/**
	print_help prints the help information for this application
*/
void print_help() {
	printf("%4s %-10s %s\n", "-w", "<value>", "warmup time in seconds (default 5)");
	printf("%4s %-10s %s\n", "-s", "<value>", "square size for smoothing (default 9)");
	printf("%4s %-10s %s\n", "-b", "<value>", "background blend (default 64)");
	printf("%4s %-10s %s\n", "-t", "<value>", "background subtraction threshold (default 16)");
}


int main(int argc, char *argv[]) {
	int i,j;
	
	/**
		receive packet variables
	*/
	unsigned char* rcv_type = malloc(sizeof(unsigned char));
	unsigned int* rcv_id = malloc(sizeof(unsigned int));
	unsigned char* rcv_pkt = malloc(128);
	unsigned int* rcv_pkt_length  = malloc(sizeof(unsigned int));
	
	/**
		timer variables
	*/
	struct timeb time_a, time_b;
	int time_diff;
	
	char* filename[16];
	
	/**
		parameters for background subtraction
	*/
	int thresh = 16, sqsize = 9, blend = 64;
	
	/**
		warmup time in seconds, used to adjust the correct initial exposure level.
	*/
	int warmup_time = 5;
	
	
	// process command line args
	i = 1;
	while (i < argc) {
		char* arg = argv[i];
		
		if (arg[0] == '-') {
			if (arg[1]) {
				switch(arg[1]) {
					case 's':
						if (argc > (i+1)) {
							j = atoi(argv[i+1]);
							if (j!=0) {
								sqsize = j;
								i++;
							}
						}
						break;
					case 'b':
						if (argc > (i+1)) {
							j = atoi(argv[i+1]);
							if (j!=0) {
								blend = j;
								i++;
							}
						}
						break;
					case 't':
						if (argc > (i+1)) {
							j = atoi(argv[i+1]);
							if (j!=0) {
								thresh = j;
								i++;
							}
						}
						break;
					case 'w':
						if (argc > (i+1)) {
							j = atoi(argv[i+1]);
							if (j!=0) {
								warmup_time = j;
								i++;
							}
						}
						break;
					case '-':
						print_help();
						return 0;
						break;
				}
			}
		}
		
		i++;
	}
	
	printf("thresh = %d\n",thresh);
	printf("blend = %d\n",blend);
	printf("sqsize = %d\n",sqsize);
	
	cameraboard* cb = cameraboard_open(IMG_COLOR_YUV_422, IMG_RES_QVGA, 0);
	
	camera_calibrate_exp(warmup_time, 0, 1);
	
	bgmodel *bgm;
	i = 0;
	ftime(&time_a); // init timing
	while(1) {
		if (0) { // wait for key press
			char cmd;
			read(0, &cmd, 1); // read char from std input
			if (cmd == 0x1B) break; // ESC char, exit loop
		}
		
		if (camera_capture_frame()) {
			printf("camera_get_frame error\n");
		}
		
		memset(filename, 0, sizeof(filename));
		sprintf(filename,"im%04d.jpg", i);
		
		if (i == 0) { // first frame, init bgmodel
			bgm = init_bgmodel(cb->img, cb->img_width, cb->img_height, thresh, sqsize, blend);
			write_jpeg_image(cb->img, cb->img_width, cb->img_height, filename, 0, 80);
		} else {
			update_bgmodel(bgm, cb->img);
			
			if (bgm->mbb->count > 0)
			{
				write_jpeg_image(bgm->fg, cb->img_width, cb->img_height, filename, 0, 80);
				
				// send number of bounding boxes
				rcv_pkt[0] = (bgm->mbb->count >> 8) & 0xFF;
				rcv_pkt[1] = bgm->mbb->count & 0xFF;

				cbpkt_send(128, BASESTATION_ID, rcv_pkt, 2);
			}
		}

		// echo any packets received
		if (cbpkt_rcv(rcv_type, rcv_id, rcv_pkt, rcv_pkt_length))
		{
			printf("received packet -- type = %d, srcid = %d, data = ", *rcv_type, *rcv_id);
			for(j=0;j<(*rcv_pkt_length);j++)
			{
				printf("%02X ", rcv_pkt[j]);
			}
			printf("\n");
			
			cbpkt_send(*rcv_type, *rcv_id, rcv_pkt, *rcv_pkt_length);

		}

		// calculate time per frame
		ftime(&time_b); // stop timing
		time_diff = (time_b.time - time_a.time)*1000 + (time_b.millitm - time_a.millitm);
		time_a.time = time_b.time; // reset time_a to beginning of next iteration
		time_a.millitm = time_b.millitm; // reset time_a to beginning of next iteration
		
		// print iteration time in ms
		printf("frame %4d - %d\n", i, time_diff);
		cbprintf(0,"frame %4d - %d\n", i, time_diff);
		
		i++;
	}
	
	free(bgm);
	cameraboard_close();
}
