/**
	planar 422 image formats
*/
enum {
	IMG_COLOR_YUV_422,
	IMG_COLOR_GBR_422,
};

/**
	the actual image is a center crop of the full 1280x1024 image region, not a downsampled image
*/
enum {
	IMG_RES_VGA,
	IMG_RES_QVGA,
	IMG_RES_QQVGA,
};




/**
	contains the image data and attributes
		image - contains the image in the format and size set by the camera initialization
		image_width - image width as set by the camera init
		image_height - image height as set by the camera init
		image_length - equivalent to image_width*image_height
*/
typedef struct cameraboard {
	unsigned char* img;
	int img_width;
	int img_height;
	int img_length;
} cameraboard;


/**
	turns on/off the auto exposure on the camera
		x - boolean value indicating whether auto exposure is to be turned on
*/
void camera_set_autoexp(int x);

/**
	initializes the camera with the indicated settings and allocates memory for the image.  returns a cameraboard struct which provides a pointer to the image data
		img_format - the image format as defined by the IMG_FORMAT_* constants above
		img_size - the image size as defined by the IMG_SIZE_* constants above
		auto-exposure - boolean value whether auto exposure is turned on
		warmup_time - time given to calibrate/warm up the camera exposure.  this time needs to be longer for bright outdoor settings
*/
cameraboard* cameraboard_open(int img_color0, int img_res0, int auto_exposure);

/**
	frees allocated memory and closes connections with devices
*/
void cameraboard_close();

/**
	captures a new frame.  the image is accessed through the camera_board->image pointer.  returns 0 if successful
*/
int camera_capture_frame();

/**
	turns on auto exposure for a given time period thereby letting the camera calibrate to the correct exposure
		calibration_time - time in sec for exposure calibration
		auto_exp_on - boolean value indicating whether auto exposure will be left on after calibration.  a value of 0 will turn off auto exposure and freeze the exposure settings
		verbose - prints a countdown starting from the calibration_time
*/
void camera_calibrate_exp(int calibration_time, int auto_exp_on, int verbose);

/**
 * enque to outgoing que.
 * return 0 if succeed, otherwise 1.
 * 	data - memory buffer which contains data
 * 	size - data size
 *      id - packet type
 *      cv_struct_ptr - pointer to a condition variable structure (or NULL if none)
 */
int cbpkt_send(unsigned char type, unsigned char* payload, unsigned int length);


/**
 * deque to incoming que.
 * return 0 if succeed, otherwise 1.
 * 	data - memory buffer which will contain data
 * 	size - pointer to memory which will contain data size
 */
int cbpkt_rcv(unsigned char* type, unsigned int* src_id_actual, unsigned char* payload, unsigned int* length);

/**
 * check if outging que is full.
 * return 1 if it's full, otherwise 0.
 */
int outqueue_is_full();

/**
 * check if incoming que is empty.
 * return 1 if it's empty, otherwise 0.
 */
int inqueue_is_empty();

/**
 * Get global time.
 * return global time if it's synchronized, otherwise return 0.
 */
unsigned long cameraboard_get_time();

ssize_t uart_writeTOSpackets(unsigned char* payload, int length, int id);

int wireless_printf(int group, unsigned char* s, int length);
