/*
	Sample code implementing a simple motion detection algorithm using frame differencing to detect
	motion between frames.

	Author: Edgar J. Lobaton (lobaton@eecs.berkeley.edu)
	Date: 11/09/2008
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
//#include <sys/timeb.h>
#include <string.h>
#include "ippi.h"
#include "cameraboard.h"

// Global variables storing the dimensions of the images
int IMAGEWIDTH;
int IMAGEHEIGHT;

// Threshold used for image differencing
#define DIFFTHRESH 20

// Global buffer used for computing image differencing.
unsigned char* MASK;


/*  Function writing an image 'im' with the specified dimensions to a file with specified
		prefix and number image count. */
void writeFileBinary(unsigned char* im, int image_width, int image_height, char* prefix, int count)
{
	int i, j;
	int image_length = image_width * image_height;
	char fn[20];
	
	printf("Writing %s%04d\n", prefix,count);
	
	// Opening file
	sprintf(fn,"%s%04d", prefix, count);
	FILE* fp = fopen(fn,"w");
	
	// Writing into file
	for (i=0; i<image_length; i++) {
			fputc(im[i], fp);
	}

	// Closing file
	fclose(fp);
	return;
}


/*  This function counts the number of pixels values that are significantly different
		between 'im0' and 'im1'. */
int frameDiff(unsigned char* im0, unsigned char* im1)
{
	int i;
	int diff;

	// Computing a mask based on frame differencing
	for(i=0; i<IMAGEHEIGHT*IMAGEWIDTH; i++) {
		diff = abs( (int)im0[i] - im1[i] );
		MASK[i] = diff > DIFFTHRESH;
	}

	// Performing Median Filtering
	int stepSize = IMAGEWIDTH*sizeof(unsigned char);
	IppiSize srcRoi = {IMAGEWIDTH,IMAGEHEIGHT};
	IppiSize mask = {9, 9};
	IppiPoint anchor = {4, 4};
	ippiFilterMedian_8u_C1R (MASK, stepSize, MASK, stepSize, srcRoi, mask, anchor);

	// Counting the number of pixels that were detected
	int val = 0;
	for(i = 0; i<IMAGEHEIGHT*IMAGEWIDTH; i++) {
		val += MASK[i];
	}

	return val;
}


int main(int argc, char *argv[]) {
	int ct = 0;
	unsigned char pkt[100];

	// Starting camera frame capturing
	int auto_exposure = 0;
	int warmup_time = 5;
	cameraboard* cb = cameraboard_open(IMG_COLOR_YUV_422, IMG_RES_QVGA, auto_exposure, warmup_time);
	camera_calibrate_exp(warmup_time, auto_exposure, 1);

	// Getting size of images
	IMAGEWIDTH = cb->img_width;
	IMAGEHEIGHT = cb->img_height;

	// Allocating space to the image that will be used for computing frame differencing
	MASK = (unsigned char*) malloc(IMAGEWIDTH*IMAGEHEIGHT*sizeof(unsigned char));

	// Getting global time ... Waits until mote is synchronized.
	//long globaltime = 0;
	//while(!(globaltime=GetGlobalTime()));

	// Printing first frame
	unsigned char* im0 = (unsigned char*) malloc(IMAGEWIDTH*IMAGEHEIGHT*sizeof(unsigned char));
	unsigned char* im1 = (unsigned char*) malloc(IMAGEWIDTH*IMAGEHEIGHT*sizeof(unsigned char));
	unsigned char* tmp;
	writeFileBinary(im0, IMAGEWIDTH, IMAGEHEIGHT, "im", 0);

	while(1) {
		// Capturing next frame and copying it to im1
		if(camera_capture_frame()) {
			printf("ERROR: camera_get_frame\n");
		}
		memcpy(im1,cb->img,IMAGEWIDTH*IMAGEHEIGHT*sizeof(unsigned char));

		// Obtaining count of number of significantly different pixel values between frames
		ct = frameDiff(im0,im1);

		// Displaying and sending results if change is significant
		if(ct>0) {
			printf("Count: %d\n",ct);

			// Creating the packet
			pkt[0] = ct/256;
			pkt[1] = ct%256;

			// Sending packet.
			int pktType = 128; // We are using the first user-available data packet type
			int pktSz = 2; // Size of the packet
			//writePkt(pktType,moteId,pkt,pktSz);
			cbpkt_send(pktType, pkt, pktSz);
		}

		// Swapping frames. The new frame becomes the old frame.
		tmp = im0;
		im0 = im1;
		im1 = tmp;
	}
	
	// Closing camera board
	cameraboard_close();
	return 0;
}
