#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/timeb.h>
#include <linux/types.h>
#include <pthread.h>

#include "camera_board.h"
#include "camera_regs.h"
#include "videodev.h"
#include "pxa_camera.h"

#include "CameraPacket.h"

#define CAMERA_I2C_ADDR 0x30
#define CAMERA_MIDH 0x1c
#define CAMERA_MIDL 0x1d
#define CAMERA_PIDH 0x0a
#define CAMERA_PIDL 0x0b
#define CHIP_DELAY  0xFF
#define I2C_SLAVE 0x0703
#define BAUDRATE B115200
//#define BAUDRATE B9600

#define YCbCrtoR(Y,Cb,Cr) (1000*Y + 1371*(Cr-128))/1000
#define YCbCrtoG(Y,Cb,Cr) (1000*Y - 336*(Cb-128) - 698*(Cr-128))/1000
#define YCbCrtoB(Y,Cb,Cr) (1000*Y + 1732*(Cb-128))/1000
#define min(x1, x2)   (((x1)<(x2))?(x1):(x2))

#define FRAME_ID_INDEX 9

// image format array - corresponds with #defines in camera_board.h
static struct {
	char *name;
	int format;
} optional_image_format[] =
{
	{"IMG_FORMAT_YUV_422", 0,},
};

// image size array - corresponds with #defines in camera_board.h
static struct {
	char *name;
	unsigned short width;
	unsigned short height;
}	optional_image_size[] = 
{
	{"VGA", 640, 480,},
	{"QVGA", 320, 240,},
	{"QQVGA", 160, 120,},
	{"SVGA", 800, 600,},
};

struct reg_set_s {
  int val1;
  int val2;
};

/*
*		global vars
*/

static int fd_i2c;
static int fd_camera;
static struct video_window vw;
static struct reg_set_s reg;
static __u8 *buf = NULL;
static struct termios tio_camera;
static fd_set rfds;
static struct timeval tv;
static unsigned short image_width;
static unsigned short image_height;
static unsigned short image_format;

static int fd_uart; // UART port
static struct termios tio_uart; // terminal settings struct

static camera_board* cb;



void delay_ms(long ms) {
	fd_set rfds;
	struct timeval tv;

	FD_ZERO(&rfds);
	FD_SET(0, &rfds);

	tv.tv_sec = 0;
	tv.tv_usec = ms*1000;

	select(1, &rfds, NULL,  NULL, &tv);
}

int i2c_write(int fd, __u8 reg, __u8 val) {
	int retries;
	__u8 data[2];

	data[0] = reg;
	data[1] = val;
	for(retries=5; retries; retries--) {
		if(write(fd, data, 2)==2)
			return 0;
		delay_ms(2);
	}
	
	return -1;
}

int i2c_read(int fd, __u8 reg, __u8 *val) {
	int retries;

	for(retries=5; retries; retries--) {
		if(write(fd, &reg, 1)==1)
			if(read(fd, val, 1)==1)
				return 0;
		delay_ms(2);
	}
	
	return -1;
}

int camera_i2c_init(int img_format, int img_size, int auto_exposure) {
	int i;
	__u8 id[4];
	__u8 val;
	int num_camregs = sizeof(camera_regs_write);

	// init camera registers, apply camera settings
	init_camera_regs();
	
	//camera_regs[CAMREG_CLKRC] = 0x01;
	//camera_regs[CAMREG_DBLV] = 0x4A; //added by leon 20080707
	
	if (auto_exposure) {
		camera_regs[CAMREG_COM8] = 0x8F; // default settings, AEC AGC AWB on
	} else {
		camera_regs[CAMREG_COM8] = 0x88; // turn off AEC AGC AWB
	}
	switch(img_format) {
		case IMG_FORMAT_YUV_422:
			camera_regs[CAMREG_COM7] = 0x02;
			break;
		case IMG_FORMAT_GRB_422:
			camera_regs[CAMREG_COM7] = 0x03;
			break;
		default: // IMG_FORMAT_YUV_422
			camera_regs[CAMREG_COM7] = 0x02;
	}
	switch(img_size) {
		case IMG_SIZE_VGA:
			camera_regs[CAMREG_HSTART] = 0x42;
			camera_regs[CAMREG_HSTOP] = 0x92;
			camera_regs[CAMREG_VSTART] = 0x23;
			camera_regs[CAMREG_VSTOP] = 0x5F;
			break;
		case IMG_SIZE_QVGA:
			camera_regs[CAMREG_HSTART] = 0x56;
			camera_regs[CAMREG_HSTOP] = 0x7E;
			camera_regs[CAMREG_VSTART] = 0x32;
			camera_regs[CAMREG_VSTOP] = 0x50;
			break;
		case IMG_SIZE_QQVGA: // actual image higher by 8 pixels from center, since HREF VREF should be different from default
			camera_regs[CAMREG_HSTART] = 0x60;
			camera_regs[CAMREG_HSTOP] = 0x74;
			camera_regs[CAMREG_VSTART] = 0x39;
			camera_regs[CAMREG_VSTOP] = 0x48;
			break;
		case IMG_SIZE_SVGA:
			camera_regs[CAMREG_HSTART] = 0x32;
			camera_regs[CAMREG_HSTOP] = 0xBD;
			camera_regs[CAMREG_VSTART] = 0x26;
			camera_regs[CAMREG_VSTOP] = 0xB1;
			break;
		default: // IMG_SIZE_640_480
			camera_regs[CAMREG_HSTART] = 0x42;
			camera_regs[CAMREG_HSTOP] = 0x92;
			camera_regs[CAMREG_VSTART] = 0x23;
			camera_regs[CAMREG_VSTOP] = 0x5F;
	}
	
#if TEST_DEBUG
	printf("open i2c device...\n");
#endif

	// open camera device
	fd_i2c = open("/dev/i2c/0", O_RDWR);
	if(fd_i2c<0) {
		fd_i2c = open("/dev/misc/i2c", O_RDWR);
		if(fd_i2c<0) {
			printf("fail to open i2c adapter device!\n");
			return -1;
		}
	}

	if(ioctl(fd_i2c, I2C_SLAVE, CAMERA_I2C_ADDR)<0) {
		printf("fail to set i2c device slave address!\n");
		close(fd_i2c);
		return -1;
	}
	
#if TEST_DEBUG
	printf("set slave address to 0x%x success!\n", CAMERA_I2C_ADDR);
#endif

#if TEST_DEBUG
	if(i2c_read(fd_i2c, CAMERA_MIDH, id)||i2c_read(fd_i2c, CAMERA_MIDL, id+1)) {
		printf("fail to get camera MID!\n");
		close(fd_i2c);
		return -1;
	}
	printf("manufactory ID is 0x%04x\n", (id[0]<<8)|id[1]);

	i2c_read(fd_i2c, CAMERA_PIDH, id+2);
	i2c_read(fd_i2c, CAMERA_PIDL, id+3);
	printf("product ID is 0x%04x\n", (id[2]<<8)|id[3]);
#endif

	// reset camera settings
	if(i2c_write(fd_i2c, 0x12, 0x80)) {
			printf("write subaddr 0x12 fail!\n");
	}
	delay_ms(20); // delay 20 ms

	// write camera register settings
	for(i=0; i<num_camregs; i++) {
		if(i2c_write(fd_i2c, camera_regs_write[i], camera_regs[camera_regs_write[i]])) {
			printf("write subaddr 0x%x fail!\n", camera_regs_write[i]);
		}
		
#if TEST_DEBUG
i2c_read(fd, camera_regs_write[i], &val);
printf("subaddr 0x%x is %x\n", camera_regs_write[i], val);
#endif

	}

	return 0;
}

__u32 Conv_YCbCr_Rgb(__u8 y0, __u8 y1, __u8 cb0, __u8 cr0) {
	int r0, g0, b0, r1, g1, b1;
	__u32 rgb0, rgb1;
	__u32 rgb;

	/* 4 frames/s @192MHz, 12MHz ; 6 frames/s @450MHz, 12MHz */
  r0 = YCbCrtoR(y0, cb0, cr0);
  g0 = YCbCrtoG(y0, cb0, cr0);
  b0 = YCbCrtoB(y0, cb0, cr0);
  r1 = YCbCrtoR(y1, cb0, cr0);
  g1 = YCbCrtoG(y1, cb0, cr0);
  b1 = YCbCrtoB(y1, cb0, cr0);

  if (r0>255 ) r0 = 255;
  if (r0<0) r0 = 0;
  if (g0>255 ) g0 = 255;
  if (g0<0) g0 = 0;
  if (b0>255 ) b0 = 255;
  if (b0<0) b0 = 0;

  if (r1>255 ) r1 = 255;
  if (r1<0) r1 = 0;
  if (g1>255 ) g1 = 255;
  if (g1<0) g1 = 0;
  if (b1>255 ) b1 = 255;
  if (b1<0) b1 = 0;


	/* 8:8:8 24bit format */
	rgb0 = (((__u8)r0)<<16) | (((__u8)g0)<<8) | (((__u8)b0)<<0);
	rgb = rgb0;

  return(rgb);
}


void PrintTime()
{
	static unsigned long last;
	unsigned long current;
	struct timeb temp;

	ftime(&temp);

	current = temp.time * 1000;
	current += temp.millitm;

	printf("delay : %lu\n", current - last);
	
	last = current;
}

unsigned char g_inque[MAX_QUE_SIZE][MAX_ENTRY_SIZE];
unsigned char g_outque[MAX_QUE_SIZE][MAX_ENTRY_SIZE];
unsigned char g_outque_id[MAX_QUE_SIZE];
cv_struct_t* g_outque_cv_struct_ptr[MAX_QUE_SIZE];

unsigned short g_inque_start, g_inque_end, g_inque_count;
unsigned short g_outque_start, g_outque_end, g_outque_count;

pthread_mutex_t g_inque_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t g_outque_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t g_time_mutex = PTHREAD_MUTEX_INITIALIZER;

unsigned long g_global_time = 0;
unsigned long g_system_time = 0;

unsigned short g_mote_id = 0xFFFF;

int EnqueIncoming(unsigned char* data, unsigned char size)
{
	if(IsInqueFull() || size > MAX_ENTRY_SIZE)
		return 1;

	pthread_mutex_lock(&g_inque_mutex);

	*(g_inque[g_inque_end]) = size;
	memcpy(g_inque[g_inque_end] + 1, data, size);

	g_inque_end = (g_inque_end + 1) % MAX_QUE_SIZE;

	g_inque_count++;

	pthread_mutex_unlock(&g_inque_mutex);

	return 0;
}

int EnqueOutgoing(unsigned char* data, unsigned char size, unsigned char id,
		  cv_struct_t* cv_struct_ptr)
{
	if(IsOutqueFull() || size > MAX_ENTRY_SIZE)
		return 1;

	pthread_mutex_lock(&g_outque_mutex);

	*(g_outque[g_outque_end]) = size;
	memcpy(g_outque[g_outque_end] + 1, data, size);
	g_outque_id[g_outque_end] = id;

	// condition variables for signalling a waiting thread
	// that the packet has been sent
	g_outque_cv_struct_ptr[g_outque_end] = cv_struct_ptr;

	g_outque_end = (g_outque_end + 1) % MAX_QUE_SIZE;

	g_outque_count++;

	pthread_mutex_unlock(&g_outque_mutex);

	return 0;
}

int DequeIncoming(unsigned char* data, unsigned char* size)
{
	if(IsInqueEmpty() || size == NULL || data == NULL)
		return 1;

	pthread_mutex_lock(&g_inque_mutex);

	*size = *(g_inque[g_inque_start]);

	memcpy(data, g_inque[g_inque_start] + 1, *size);

	g_inque_start = (g_inque_start + 1) % MAX_QUE_SIZE;

	g_inque_count--;

	pthread_mutex_unlock(&g_inque_mutex);

	return 0;
}

int DequeOutgoing(unsigned char* data, unsigned char* size, unsigned char* id)
{
	if(IsOutqueEmpty() || size == NULL || data == NULL)
		return 1;

	pthread_mutex_lock(&g_outque_mutex);

	*size = *(g_outque[g_outque_start]);
	*id = g_outque_id[g_outque_start];
	memcpy(data, g_outque[g_outque_start] + 1, *size);

	// condition variables for signalling a waiting thread
	// that the packet has been sent
	if (g_outque_cv_struct_ptr[g_outque_start] != NULL) {
	  pthread_mutex_lock(g_outque_cv_struct_ptr[g_outque_start]->mutex_ptr);
	  pthread_cond_signal(g_outque_cv_struct_ptr[g_outque_start]->cv_ptr);
	  pthread_mutex_unlock(g_outque_cv_struct_ptr[g_outque_start]->mutex_ptr);
	}

	g_outque_start = (g_outque_start + 1) % MAX_QUE_SIZE;

	g_outque_count--;

	pthread_mutex_unlock(&g_outque_mutex);

	return 0;
}

int IsInqueFull()
{
//	return ((g_inque_end + 1) % MAX_QUE_SIZE == g_inque_start);
	return (g_inque_count == MAX_QUE_SIZE);
}
int IsOutqueFull()
{
//	return ((g_outque_end + 1) % MAX_QUE_SIZE == g_outque_start);
	return (g_outque_count == MAX_QUE_SIZE);
}
int IsInqueEmpty()
{
//	return (g_inque_start == g_inque_end);
	return (g_inque_count == 0);
}
int IsOutqueEmpty()
{
//	return (g_outque_start == g_outque_end);
	return (g_outque_count == 0);
}

int GetInqueFreeBytes()
{
	return (MAX_QUE_SIZE - g_inque_count);
}

int GetOutqueFreeBytes()
{
	return (MAX_QUE_SIZE - g_inque_count);
}

void InitQue()
{
	g_inque_start = g_inque_end = g_outque_start = g_outque_end = 0;
	g_inque_count = g_outque_count = 0;

	
}

unsigned long GetGlobalTime()
{
	struct timeval temp_time;
	unsigned long local;

	if(g_global_time == 0)
		return 0;

	pthread_mutex_lock(&g_time_mutex);

	gettimeofday(&temp_time, NULL);
	local = temp_time.tv_sec * 1000;
	local += temp_time.tv_usec / 1000;

	local = local - g_system_time + g_global_time; 

	pthread_mutex_unlock(&g_time_mutex);

	return local;
}

void* receive_thread(void* arg) {
	unsigned char c, c_prev;
	unsigned char in_data[MAX_ENTRY_SIZE];
	unsigned char in_length;
	unsigned char in_id;
	unsigned long in_drop = 0;
	unsigned char out_data[MAX_ENTRY_SIZE];
	unsigned char out_length = 1;
	unsigned char out_id = PACKET_TYPE_NOOUTGOING;
	unsigned long out_drop = 0;
	unsigned long time_comp;
	int write_len;
	int i;
	struct timeval temp_time;

	int errcount = 0;

	unsigned char data[100];

	for(i=0;i<100;i++)
		data[i] = i+1;

	while(1) 
	{
		errcount = 0;
		while(1) 
		{
			uart_read(&c, 1);
			if ((c == 0x7E))
				break;	
		}

		in_length = 0;

		while(1) 
		{
			uart_read(&c, 1);
			if (c == 0x7E) 
			{
				break;
			} 
			else if (c_prev == 125) 
			{
				if (c == 93) 
				{
					in_data[in_length-1] = 125;
				} 
				else if (c == 94) {
					in_data[in_length-1] = 126;
				}
			} 
			else 
			{
				in_data[in_length] = c;
				in_length++;
			}
			c_prev = c;
		}

/*
		for(i=0;i<10;i++)
		{
			printf("%02X ", in_data[i]);
		}
		printf("\n");
*/

		in_id = in_data[FRAME_ID_INDEX];

		if(in_id == PACKET_TYPE_NOINCOMING)
		{
			//ok signal, do nothing.
		}	
		else if(in_id == PACKET_TYPE_CAMERASYNC)
		{
			//global time sync
			gettimeofday(&temp_time, NULL);

			pthread_mutex_lock(&g_time_mutex);

			g_system_time = temp_time.tv_sec * 1000;
			g_system_time += temp_time.tv_usec / 1000;
		
			g_global_time = in_data[FRAME_ID_INDEX+1] << 24;
			g_global_time += in_data[FRAME_ID_INDEX+2] << 16;
			g_global_time += in_data[FRAME_ID_INDEX+3] << 8;
			g_global_time += in_data[FRAME_ID_INDEX+4];

			pthread_mutex_unlock(&g_time_mutex);

			g_mote_id = in_data[FRAME_ID_INDEX+5] << 8;
			g_mote_id += in_data[FRAME_ID_INDEX+6];

			printf("camera sync : global (%lu), local (%lu), moteid(%d)\n", g_global_time, g_system_time, g_mote_id);

			//for Backbround img construction.
			EnqueIncoming(in_data+3, in_length-3);
		}
		else if(in_id == PACKET_TYPE_TIMECOMP)
		{
printf("timecomp packet\n");
			//time compansation
			time_comp = in_data[FRAME_ID_INDEX+1] << 24;
			time_comp += in_data[FRAME_ID_INDEX+2] << 16;
			time_comp += in_data[FRAME_ID_INDEX+3] << 8;
			time_comp += in_data[FRAME_ID_INDEX+4];

			pthread_mutex_lock(&g_time_mutex);
			
			g_global_time += time_comp;
	
			pthread_mutex_unlock(&g_time_mutex);
			printf("global comp time : %lu\n", time_comp);
		}
		else if(in_id == PACKET_TYPE_RETRANSREQ)
		{
			//printf("retrans\n");
			//nak, data retransmission
		}
		else
		{
			//cut out MSG_TYPE and SEQ of serial communication.
			if( EnqueIncoming(in_data+3, in_length-3) )
			{
				printf("Packet drop from recv thread!!\n");
				in_drop++;
			}
		}

/*
		if(in_id == PACKET_TYPE_CAMERASYNC)
		{
			//for global time compansation
			out_length = 1;
			out_id = PACKET_TYPE_TIMEREFLECT;
		}
		else if(in_id == PACKET_TYPE_RETRANSREQ)
*/
		if(in_id == PACKET_TYPE_RETRANSREQ)
		{
			//for data retransmisson
			printf("retrans id(%d), len(%d)\n", out_id, out_length);
		}
		else
		{
			if( DequeOutgoing(out_data, &out_length, &out_id) == 0)
			{
				//have normal data to send
			}
			else
			{
				//have nothing to send
				out_length = 1;
				out_id = PACKET_TYPE_NOOUTGOING;
			}
		}

		write_len = uart_writeTOSpackets(out_data, out_length, out_id);

	}
}

/* * * * * * *
*
*		API functions start here
*
* * * * * * */

void camera_set_auto_exp(int x) {
	if (x) {
		if(i2c_write(fd_i2c, CAMREG_COM8, 0x8F)) printf("set auto exp fail!");
	} else {
		if(i2c_write(fd_i2c, CAMREG_COM8, 0x88)) printf("set auto exp fail!");
	}
}

camera_board* camera_board_init(int img_format, int img_size, int auto_exposure, int warmup_time) {
	cb = malloc(sizeof(camera_board));
	
/*
*		UART inits
*/

	// open UART port
	fd_uart = open("/dev/tts/2", O_RDWR | O_NOCTTY ); 
	if (fd_uart <0) {
		printf("Error in opening UART\n");
		exit(-1); 
	}

	bzero(&tio_uart, sizeof(tio_uart));

	// set com port parameters
	tio_uart.c_cflag = BAUDRATE | CRTSCTS | CS8 | CLOCAL | CREAD;
	tio_uart.c_iflag = IGNPAR;
	tio_uart.c_oflag = 0;

	// set input mode (non-canonical, no echo,...)
	tio_uart.c_lflag = 0;

	tio_uart.c_cc[VTIME] = 0;   // inter-character timer unused
	tio_uart.c_cc[VMIN] = 1;   // blocking read until 1 chars received

	// clear port     
	tcflush(fd_uart, TCIFLUSH);

	// activate new settings
	tcsetattr(fd_uart,TCSANOW,&tio_uart);
	
/*
*		camera inits
*/


	fd_camera = open("/dev/v4l/video0", O_RDONLY); //rd&wr
  if(fd_camera<0)
  {
    fprintf(stderr, "Open camera fail!\n");
  }
  else
  {
    fprintf(stdout, "Open camera success\n");
  }

	image_width = optional_image_size[img_size].width;
  image_height = optional_image_size[img_size].height;
  image_format = optional_image_format[img_format].format;

#if TEST_DEBUG
	printf("width:%d  height:%d  format:%d\n", image_width, image_height, image_format);
	printf("setting image size\n");
#endif

	/* setting image size */
	vw.width = image_width;
	vw.height = image_height;

	if(ioctl(fd_camera, VIDIOCSWIN, &vw)<0)
    printf("VIDIOCSWIN fail\n");

#if TEST_DEBUG
  if(ioctl(fd_camera, VIDIOCGWIN, &vw)<0)
    printf("VIDIOCGWIN fail\n");
  else
    printf("current width %d, height %d\n", vw.width, vw.height);
#endif

	if(camera_i2c_init(img_format, img_size, auto_exposure))
  {
    printf("fail to initialize the I2C interface of camera!\n");
  }

	/* setting camera image format */
	reg.val1 = CAMERA_IMAGE_FORMAT_YCBCR422_PACKED;
	reg.val2 = CAMERA_IMAGE_FORMAT_YCBCR422_PLANAR;

	if(ioctl(fd_camera, WCAM_VIDIOCSINFOR, &reg)<0)
      printf("WCAM_VIDIOCSINFOR fail\n");

#if TEST_DEBUG
  if(ioctl(fd_camera, WCAM_VIDIOCGINFOR, &reg)<0)
    printf("WCAM_VIDIOCGINFOR fail\n");
  else
    printf("current input format %d, output format %d\n", reg.val1, reg.val2);
#endif

#if TEST_DEBUG
	puts("allocate memory for camera buffer.\n");
#endif

  buf = malloc(image_width*image_height*2);
	
  if(!buf)
	{
	  printf("fail to allocate memory for camera!\n");
		close(fd_camera);
	}

#if TEST_DEBUG
	printf("buffer at 0x%08x\n", (int)buf);
#endif

	if(tcgetattr(0, &tio_camera)<0) {
    printf("tcgetattr fail!\n");
//    goto fail;
  } else {
    tio_camera.c_lflag &= ~(ICANON | ECHO);
    if(tcsetattr(0, TCSAFLUSH, &tio_camera)<0) {
      printf("tcsetattr fail!\n");
//      goto fail;
    }
  }

	/* setting to get a frame */
	if(ioctl(fd_camera, VIDIOCCAPTURE, STILL_IMAGE)<0)
	{
		printf("VIDIOCCAPTURE fail\n");
//		goto fail;
	}

	FD_ZERO(&rfds);
	
/*
 * que init
 */

	InitQue();


	
/*
*		camera board struct inits
*/
	
	cb->image = buf;
	cb->image_width = image_width;
	cb->image_height = image_height;
	cb->image_length = image_width * image_height;
	
	return cb;
}

void camera_board_start_receiving()
{
/*
*		receive thread inits
*/
	pthread_t thread_rcv;
	int pthread_errorcode;
	if (pthread_errorcode = pthread_create(&thread_rcv, NULL, receive_thread, 0)) {
		printf("error creating receive thread %d\n", pthread_errorcode);
	}

}

int camera_board_is_sync()
{
	return (g_mote_id != 0xFFFF);
}

void camera_board_close() {
	free(buf);
	free(cb);
	close(fd_i2c);
	close(fd_camera);
	close(fd_uart);
}

int camera_get_frame() {
	int i;

	FD_SET(0, &rfds);
	FD_SET(fd_camera, &rfds);

	tv.tv_sec = 3;
	tv.tv_usec = 0;

	select(fd_camera+1, &rfds, NULL,  NULL, &tv);

	if(FD_ISSET(fd_camera, &rfds)) {	
		i = read(fd_camera, buf, image_width*image_height*2);

		if(i<0) {
			fprintf(stderr, "read fail! %d\n", i);
		}
	}
	
	return 0;
}

void camera_calibrate_exposure(int calibration_time, int auto_exp_on, int verbose) {
	int i,j;
	struct timeb time_a;
	
	if (verbose) printf("calibrating exposure\n");
	
	camera_set_auto_exp(1); // turn on auto exposure
	
	ftime(&time_a);
	j = time_a.time;
	for(i=0; i<calibration_time; i++) {
		if (verbose) printf("%d\n",(calibration_time-i));
		while(time_a.time == j) ftime(&time_a); // wait till next second
		j = time_a.time;
		camera_get_frame();
	}
	
	if (!auto_exp_on) camera_set_auto_exp(0); // turn off auto exposure
	
	if (verbose) printf("done\n");
}

ssize_t uart_read(unsigned char* packet, int length) {
	return read(fd_uart, packet, length);
}

ssize_t uart_write(unsigned char* packet, int length) {
	return write(fd_uart, packet, length);
}

ssize_t uart_writeTOSpackets(unsigned char* payload, int length, int id) {
	int i,j;
	unsigned char tospacket[256];
	int tospacketindex;
	int payloadindex;
	ssize_t ret;
	unsigned short crc;

	tospacket[0]=0x7e;
	tospacket[1]=0x45;
	tospacket[2]=0x0e;
	tospacket[3]=0x00;
	tospacket[4]=0x0;
	tospacket[5]=0x0;
	tospacket[6]=0x0;
	tospacket[7]=0x0;
	tospacket[8]=(unsigned char)length+2;
	tospacket[9]=0x00;
	tospacket[10]=(unsigned char)id; // handler ID must = 0x64

	tospacketindex = 11;

	crc = 0;

	for(i=0;i<length;i++)
		crc = crc_byte(crc, payload[i]);

	if( (crc & 0xff00) == 0x7E00)
	{
		tospacket[tospacketindex++] = 0x7D;
		tospacket[tospacketindex++] = 0x5E;
	}
	else if( (crc & 0xff00) == 0x7D00)
	{
		tospacket[tospacketindex++] = 0x7D;
		tospacket[tospacketindex++] = 0x5D;
	}
	else
		tospacket[tospacketindex++] = (crc >> 8) & 0xff;

	if( (crc & 0xff) == 0x7E)
	{
		tospacket[tospacketindex++] = 0x7D;
		tospacket[tospacketindex++] = 0x5E;
	}
	else if( (crc & 0xff) == 0x7D)
	{
		tospacket[tospacketindex++] = 0x7D;
		tospacket[tospacketindex++] = 0x5D;
	}
	else
		tospacket[tospacketindex++] = crc & 0xff;

	for(i=0;i<length;i++)
	{
		if (payload[i] == 0x7e) 
		{
      tospacket[tospacketindex++] = 0x7d;
      tospacket[tospacketindex++] = 0x5e;
    } 
		else if (payload[i] == 0x7d) 
		{
      tospacket[tospacketindex++] = 0x7d;
      tospacket[tospacketindex++] = 0x5d;
    }
		else 
		{
      tospacket[tospacketindex++] = payload[i];
    }
	}	

	//printf("len %d , payload %d bytes send with CRC %02X %02X \n", tospacketindex, length, tospacket[tospacketindex-2], tospacket[tospacketindex-1]);

	// tos packet footer
	tospacket[tospacketindex++] = 0x00;
	tospacket[tospacketindex++] = 0x00;
	tospacket[tospacketindex++] = 0x7e;

	ret = uart_write(tospacket, tospacketindex);

/*
	if(id != PACKET_TYPE_NOOUTGOING)
		printf("uart data send id %d\n", id);
*/

	return ret;
}

unsigned short crc_byte(unsigned short crc, unsigned char b)
{
  unsigned short i;

  crc = crc ^ b << 8;
  i = 8;
  do
    if (crc & 0x8000)
      crc = crc << 1 ^ 0x1021;
    else
      crc = crc << 1;
  while (--i);

  return crc;
}


unsigned short get_pkt_src_address(unsigned char* pkt)
{
	unsigned short addr;

	addr = pkt[0] << 8;
	addr += pkt[1];

	return addr;
}

unsigned char get_pkt_length(unsigned char* pkt)
{
	return pkt[4];
}

unsigned char get_pkt_type(unsigned char* pkt)
{
	return pkt[6];
}

unsigned char* get_pkt_payload(unsigned char* pkt)
{
	return pkt + 7;
}
