/**
	planar 422 image formats
*/
#define IMG_FORMAT_YUV_422		0
#define IMG_FORMAT_GRB_422		1

/**
	the actual image is a center crop of the full 1280x1024 image region, not a downsampled image
*/
#define IMG_SIZE_VGA			0
#define IMG_SIZE_QVGA			1
#define IMG_SIZE_QQVGA		2
#define IMG_SIZE_SVGA			3

/**
	buffer size (in bytes) for incoming messages
*/
#define TOSPACKET_CONTAINER_SIZE	256

/**
	input and output que size
*/
#define MAX_QUE_SIZE	4096 
#define MAX_ENTRY_SIZE	128




/**
	contains the image data and attributes
		image - contains the image in the format and size set by the camera initialization
		image_width - image width as set by the camera init
		image_height - image height as set by the camera init
		image_length - equivalent to image_width*image_height
*/
typedef struct camera_board {
	unsigned char* image;
	int image_width;
	int image_height;
	int image_length;
} camera_board;

/**
	contains a pair of pointers to a condition variable and a mutex for signalling when a packet has been sent
		cv_ptr - pointer to the condition variable
		mutex_ptr - pointer to the mutex (this is to conform with cv usage semantics)
 *
 *      Note that the purpose of the condition variables is for
 *      signaling threads that are blocking until a packet is about to
 *      be sent over the UART.  Thus, a typical usage scenario is:
 * 	  pthread_mutex_lock(&bgsend_mutex);
 *           ...
 * 	     EnqueOutgoing(packet, temp_len+6, PACKET_TYPE_BACKGROUND, bgsend_cv_struct);
 *           ...
 *	  pthread_cond_wait(&bgsend_cv,&bgsend_mutex);
 *	  pthread_mutex_unlock(&bgsend_mutex);
 *      Here, we do not need to "retest a condition" (Mesa-semantics
 *      for CV), since we are just interested in the signal to wake up
 *      the thread, not in any shared variables.
 */
typedef struct cv_struct {
  pthread_cond_t* cv_ptr;
  pthread_mutex_t* mutex_ptr;
} cv_struct_t;

/**
	turns on/off the auto exposure on the camera
		x - boolean value indicating whether auto exposure is to be turned on
*/
void camera_set_auto_exp(int x);

/**
	initializes the camera with the indicated settings and allocates memory for the image.  returns a camera_board struct which provides a pointer to the image data
		img_format - the image format as defined by the IMG_FORMAT_* constants above
		img_size - the image size as defined by the IMG_SIZE_* constants above
		auto-exposure - boolean value whether auto exposure is turned on
		warmup_time - time given to calibrate/warm up the camera exposure.  this time needs to be longer for bright outdoor settings
*/
camera_board* camera_board_init(int img_format, int img_size, int auto_exposure, int warmup_time);

/**
	frees allocated memory and closes connections with devices
*/
void camera_board_close();

/**
	captures a new frame.  the image is accessed through the camera_board->image pointer.  returns 0 if successful
*/
int camera_get_frame();

/**
	turns on auto exposure for a given time period thereby letting the camera calibrate to the correct exposure
		calibration_time - time in sec for exposure calibration
		auto_exp_on - boolean value indicating whether auto exposure will be left on after calibration.  a value of 0 will turn off auto exposure and freeze the exposure settings
		verbose - prints a countdown starting from the calibration_time
*/
void camera_calibrate_exposure(int calibration_time, int auto_exp_on, int verbose);

/**
	reads data from the UART
		packet - data array to store the incoming packet
		length - length in bytes to read from UART
*/
ssize_t uart_read(unsigned char* packet, int length);

/**
	writes data to the UART
		packet - data array of bytes to be written to UART
		length - length in bytes to write from packet
*/
ssize_t uart_write(unsigned char* packet, int length);

/**
	writes a TOS packet to UART.  automatically adds the header and segments the data into multiple packets (if contents larger than one packet)
		a - data array to be written and packaged as a TOS packet(s)
		length - length in bytes to be written
*/
ssize_t uart_writeTOSpackets(unsigned char* a, int length, int id);

/**
 * for test, print excution time since past call of PrintTime()
*/
void PrintTime();

/**
 * enque to outgoing que.
 * return 0 if succeed, otherwise 1.
 * 	data - memory buffer which contains data
 * 	size - data size
 *      id - packet type
 *      cv_struct_ptr - pointer to a condition variable structure (or NULL if none)
 */
int EnqueOutgoing(unsigned char* data, unsigned char size, unsigned char id, cv_struct_t* cv_struct_ptr);


/**
 * deque to incoming que.
 * return 0 if succeed, otherwise 1.
 * 	data - memory buffer which will contain data
 * 	size - pointer to memory which will contain data size
 */
int DequeIncoming(unsigned char* data, unsigned char* size);

/**
 * check if outging que is full.
 * return 1 if it's full, otherwise 0.
 */
int IsOutqueFull();

/**
 * check if incoming que is empty.
 * return 1 if it's empty, otherwise 0.
 */
int IsInqueEmpty();

/**
 * Get global time.
 * return global time if it's synchronized, otherwise return 0.
 */
unsigned long GetGlobalTime();

/*
 * Start communication thread between camera and telos board.
 */
void camera_board_start_receiving();

/*
 * Check whether camera board has synchronized with tmote.
 */
int camera_board_is_sync();

/*
 * CRC making function.
 */
unsigned short crc_byte(unsigned short crc, unsigned char b);

/*
 * packet accessor functions.
 */
unsigned short get_pkt_src_address(unsigned char* pkt);
unsigned char get_pkt_length(unsigned char* pkt);
unsigned char get_pkt_type(unsigned char* pkt);
unsigned char* get_pkt_payload(unsigned char* pkt);
