/*
	Sample code implementing a simple motion detection algorithm using frame differencing to detect
	motion between frames.

	Author: Edgar J. Lobaton (lobaton@eecs.berkeley.edu)
	Date: 11/09/2008
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
//#include <sys/timeb.h>
#include <string.h>
#include "ippi.h"
#include "camera_board.h"
#include "CameraPacket.h"


// Dimensions of the downsampled image
#define DOWNSAMPLEWIDTH 200
#define DOWNSAMPLEHEIGHT 150
#define DOWNSAMPLELENGTH DOWNSAMPLEWIDTH*DOWNSAMPLEHEIGHT

// Some parameters used for downsampling
#define VERTSKIP 800*800/DOWNSAMPLEWIDTH
#define HORSKIP 800/DOWNSAMPLEWIDTH

// Threshold used for image differencing
#define DIFFTHRESH 20

// Global buffer used for computing image differencing.
unsigned char* MASK;


/*  Function writing an image 'im' with the specified dimensions to a file with specified
		prefix and number image count. */
void writeFileBinary(unsigned char* im, int image_width, int image_height, char* prefix, int count)
{
	int i, j;
	int image_length = image_width * image_height;
	char fn[20];
	
	printf("Writing %s%04d\n", prefix,count);
	
	// Opening file
	sprintf(fn,"%s%04d", prefix, count);
	FILE* fp = fopen(fn,"w");
	
	// Writing into file
	for (i=0; i<image_length; i++) {
			fputc(im[i], fp);
	}

	// Closing file
	fclose(fp);
	return;
}

/*  Function performing the downsample of image 'frame'. The output is placed in
		image 'newframe'.
*/
void downsample(unsigned char* frame, unsigned char* newframe)
{
	int j,k;
	
	for(j = 0; j<DOWNSAMPLEHEIGHT; j++) {
		for(k = 0; k<DOWNSAMPLEWIDTH; k++) {
			newframe[j*DOWNSAMPLEWIDTH+k] = frame[j*VERTSKIP+k*HORSKIP];
		}
	}

	return;
}

/*  This function counts the number of pixels values that are significantly different
		between 'im0' and 'im1'. */
int frameDiff(unsigned char* im0, unsigned char* im1)
{
	int i;
	int diff;

	// Computing a mask based on frame differencing
	for(i=0; i<DOWNSAMPLELENGTH; i++) {
		diff = abs( (int)im0[i] - im1[i] );
		MASK[i] = diff > DIFFTHRESH;
	}

	// Performing Median Filtering
	int stepSize = DOWNSAMPLEWIDTH*sizeof(unsigned char);
	IppiSize srcRoi = {DOWNSAMPLEWIDTH,DOWNSAMPLEHEIGHT};
	IppiSize mask = {9, 9};
	IppiPoint anchor = {4, 4};
	ippiFilterMedian_8u_C1R (MASK, stepSize, MASK, stepSize, srcRoi, mask, anchor);

	// Counting the number of pixels that were detected
	int val = 0;
	for(i = 0; i<DOWNSAMPLELENGTH; i++) {
		val += MASK[i];
	}

	return val;
}


/*  Writes a tinyos packet to UART containing payload specified in 'pkt'. The constructed
		packet follows the standard specified for our platform. See the documentation. */
void writePkt(int pktType, int moteid, unsigned char* pkt, int pktSz) {
		int i,j;
		unsigned char* cameramotemsg = malloc(AM_PAYLOAD_LENGTH);

		// Header
		cameramotemsg[0]=(unsigned char) pktType;
		cameramotemsg[1]=(unsigned char) moteid/256;
		cameramotemsg[2]=(unsigned char) moteid%256;
		cameramotemsg[3]=0x00;
		cameramotemsg[4]=0x00;
		cameramotemsg[5]=0x00;
		cameramotemsg[6]=0x00;
		cameramotemsg[7]=0x00;
		cameramotemsg[8]=0x00;
		cameramotemsg[9]=0x00;
		
		// Payload
		memcpy(cameramotemsg+3,pkt,pktSz);
		
		//uart_writeTOSpackets(cameramotemsg, pktSz+11, moteid);
		EnqueOutgoing(cameramotemsg, pktSz+3, moteid, NULL);
		free(cameramotemsg);
}


int main(int argc, char *argv[]) {
	int ct;
	unsigned char pkt[100];

	// Allocating space to the image that will be used for computing frame differencing
	MASK = (unsigned char*) malloc(DOWNSAMPLEWIDTH*DOWNSAMPLEHEIGHT*sizeof(unsigned char));

	// Starting camera frame capturing
	int res = IMG_SIZE_SVGA;
	int auto_exposure = 0;
	int warmup_time = 5;
	camera_board* cb = camera_board_init(IMG_FORMAT_YUV_422, res, auto_exposure, warmup_time);
	camera_calibrate_exposure(warmup_time, auto_exposure, 1);

	// Initiating Receive packet thread from mote
	camera_board_start_receiving();
	
	// Getting global time ... Waits until mote is synchronized.
	//long globaltime = 0;
	//while(!(globaltime=GetGlobalTime()));

	// Printing first frame
	unsigned char* im0 = (unsigned char*) malloc(DOWNSAMPLEWIDTH*DOWNSAMPLEHEIGHT*sizeof(unsigned char));
	unsigned char* im1 = (unsigned char*) malloc(DOWNSAMPLEWIDTH*DOWNSAMPLEHEIGHT*sizeof(unsigned char));
	unsigned char* tmp;
	downsample(cb->image, im0);
	writeFileBinary(im0, DOWNSAMPLEWIDTH, DOWNSAMPLEHEIGHT, "im", 0);

	while(1) {
		// Capturing next frame
		if(camera_get_frame()) {
			printf("ERROR: camera_get_frame\n");
		}
		downsample(cb->image,im1);

		// Obtaining count of number of significantly different pixel values between frames
		ct = frameDiff(im0,im1);

		// Displaying and sending results if change is significant
		if(ct>0) {
			printf("Count: %d\n",ct);

			// Creating the packet
			pkt[0] = ct/256;
			pkt[1] = ct%256;

			// Sending packet.
			int pktType = 128; // We are using the first user-available data packet type
			int moteId = 1; // Mote ID
			int pktSz = 2; // Size of the packet
			writePkt(pktType,moteId,pkt,pktSz);
		}

		// Swapping frames. The new frame becomes the old frame.
		tmp = im0;
		im0 = im1;
		im1 = tmp;
	}
	
	// Closing camera board
	camera_board_close();
	return 0;
}
